# IPC Commands are in the following format
# {2digitcode}: {arg1}|||{arg2}|||{argn}

import json
from ._types import *
from .utils import *


def show_quick_pick(items: list, options: QuickPickOptions = None) -> QuickPickItem:
    """
    Shows a selection list allowing multiple selections.

    Returns either the selected items or undefined.
    """
    data = []
    for item in items:
        if isinstance(item, QuickPickItem):
            data.append(item.__dict__)
        else:
            data.append(item)

    items = json.dumps(data)
    if not options:
        options = {}

    if isinstance(options, QuickPickOptions):
        options = options.__dict__
    options = json.dumps(options)
    print(f"QP: {items}|||{options}", flush=True, end="")
    res = json_input()
    if not res or isinstance(res, str):
        return res
    else:
        return QuickPickItem(**res)


def show_input_box(options: InputBoxOptions = None) -> str:
    """
    Opens an input box to ask the user for input.

    The returned value will be undefined if the input box was canceled (e.g. pressing ESC). Otherwise the returned value will be the string typed by the user or an empty string if the user did not type anything but dismissed the input box with OK.
    """
    if not options:
        options = {}
    if isinstance(options, InputBoxOptions):
        options = options.__dict__
    options = json.dumps(options)
    print(f"IB: {options}", flush=True, end="")
    return uinput()


def _base(func: str, text: str, *args: str) -> str:
    print(
        f"SM: {func}|||{text}" + "|||" * bool(args) + "|||".join(args),
        flush=True,
        end="",
    )
    res = uinput()
    return res


def show_info_message(text: str, *args: str) -> str:
    """
    Show an information message.
    """
    return _base("showInformationMessage", text, *args)


def show_warn_message(text: str, *args: str) -> str:
    """
    Show a warning message.
    """
    return _base("showWarningMessage", text, *args)


def show_error_message(text: str, *args: str) -> str:
    """
    Show an error message.
    """
    return _base("showErrorMessage", text, *args)


def set_status_bar_message(text: str, hide_after_timeout: int = None) -> Disposable:
    """
    Set a message to the status bar.

    Note that status bar messages stack and that they must be disposed when no longer used.

    hide_after_timeout: Timeout in seconds after which the message will be auto disposed.
    """

    print(
        f"BM: {text}"
        + f"|||{hide_after_timeout*1000}" * (hide_after_timeout is not None),
        flush=True,
        end="",
    )
    res = uinput()
    return Disposable(res)


class ActiveTextEditor:
    """
    The currently active editor or undefined. The active editor is the one that currently has focus or, when none has focus, the one that has changed input most recently.
    """

    def __init__(self):
        print("AT", flush=True, end="")
        res = uinput()
        if not res:
            self = undefined
        res = json.loads(res.replace(r"\\", r"\\\\"))
        self.__dict__.update(apply_func_to_keys(res, camel_to_snake))
        self.document = TextDocument(self.document)
        if hasattr(self, "selection"):
            self.selection = Range.from_dict(self.selection)

    def replace(self, location: Range, value: str) -> bool:
        """
        Replace a certain text region with a new value.
        You can use \r\n or \n in value and they will be normalized to the current document.
        """

        if isinstance(location, Range):
            location = location.__dict__
        location = json.dumps(location)
        print(f"EE: {location}|||{value}", flush=True, end="")
        return eval(uinput().title())

    def delete(self, location: Range) -> bool:
        """
        Delete a certain text region.
        """
        return self.replace(location, "")

    def insert(self, location: Position, value: str) -> bool:

        """
        Insert text at a location. You can use \r\n or \n in value and they will be normalized to the current document.
        """

        return self.replace(Range(location, location), value)



from ._types import *


class Extension:
    def __init__(
        self,
        name: str,
        display_name: str,
        version: str,
        description: str = None,
        icon: str = None,
        repository: dict = None,
        publisher: str = None,
    ) -> None:
        if " " in name:
            raise ValueError(
                "The Extension name should not contain any spaces! If you want to include spaces please use the display_name attribute."
            )
        self.name = name
        self.display_name = display_name
        self.version = version
        self.description = description
        self.icon = icon
        self.repository = repository
        self.publisher = publisher

        self.commands = []
        self.events = {}
        self.default_category = None
        self.keybindings = []
        self.activity_bar = None
        self.activity_bar_webview = None

    def register_command(
        self,
        func,
        name: str = None,
        title: str = None,
        category: str = None,
        keybind: str = None,
        when: str = None,
    ) -> None:
        name = func.__name__ if name is None else name
        category = self.default_category if category is None else category
        command = Command(name, func, title, category, keybind, when)
        if keybind:
            self.register_keybind(command)
        self.commands.append(command)

    def command(
        self,
        name: str = None,
        title: str = None,
        category: str = None,
        keybind: str = None,
        when: str = None,
    ):
        def decorator(func):
            self.register_command(func, name, title, category, keybind, when)
            return func

        return decorator

    def event(self, func):
        name = func.__name__.replace("on_", "")
        self.events[name] = func
        return func

    def register_keybind(self, command) -> None:
        keybind = {"command": command.extension(self.name), "key": command.keybind}
        if command.when:
            keybind.update({"when": command.when})
        self.keybindings.append(keybind)

    def set_repository(self, url: str, repo_type: str = "git") -> None:
        self.repository = {"type": repo_type, "url": url}

    def set_default_category(self, category) -> None:
        self.default_category = category

    def set_activity_bar(self, activity_bar, webview=None) -> None:
        if isinstance(activity_bar, ActivityBar):
            self.activity_bar = activity_bar.__dict__
        elif isinstance(activity_bar, dict):
            self.activity_bar = activity_bar
        else:
            raise TypeError(
                "activity_bar must be either an instance of vscode.ActivityBar or dict"
            )
        if webview:
            if isinstance(webview, StaticWebview):
                self.activity_bar_webview = webview.__dict__
            elif isinstance(webview, dict):
                self.activity_bar_webview = webview
            else:
                raise TypeError(
                    "activity_bar_webview must be either an instance of vscode.StaticWebview or dict"
                )


class Command:
    def __init__(
        self,
        name: str,
        func,
        title: str = None,
        category: str = None,
        keybind: str = None,
        when: str = None,
    ):
        self.name = self.convert_snake_to_camel(name)
        if title is None:
            self.title = self.convert_snake_to_title(name)
        else:
            self.title = title

        self.func = func
        self.func_name = self.func.__name__
        self.category = None if category is False else category
        self.keybind = keybind.upper() if keybind is not None else None
        self.when = self.convert_python_condition(when) if when is not None else None

    def extension(self, ext_name: str) -> str:
        return f"{ext_name}.{self.name}"

    def convert_snake_to_camel(self, text: str) -> str:
        temp = text.split("_")
        return temp[0] + "".join(ele.title() for ele in temp[1:])

    def convert_snake_to_title(self, text) -> str:
        return text.replace("_", " ").title()

    def convert_python_condition(self, condition) -> str:
        condition = " ".join(
            i if "_" not in i else self.convert_snake_to_camel(i)
            for i in condition.split(" ")
        )
        condition = condition.replace(" and ", " && ")
        condition = condition.replace(" or ", " || ")
        if " not " in condition:
            if not ("(" in condition and ")" in condition):
                raise SyntaxError(
                    "Use parenthesis '()' while using 'not' otherwise your conditions might not work as expected!"
                )
            else:
                condition = condition.replace(" not ", " !")

        return condition

    def __repr__(self):
        return f"<vscode.Command {self.name}>"



__author__ = "Swas.py"
__title__ = "vscode"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Swas.py"
__version__ = "1.4.6"

from . import window
from .compiler import build
from .extension import Extension
from .envMethods import env
from . import _types as ext
from ._types import *



from .utils import uinput


class Env:
    def open_external(self, target: str) -> None:
        print(f"OE: {target}", flush=True, end="")

    @property
    def app_name(self):
        print(f"EP: appName", flush=True, end="")
        return uinput()

    @property
    def app_root(self):
        print(f"EP: appRoot", flush=True, end="")
        return uinput()

    @property
    def clipboard(self):
        print(f"EP: clipboard", flush=True, end="")
        return uinput()

    @property
    def is_new_app_install(self) -> bool:
        print(f"EP: isNewAppInstall", flush=True, end="")
        return eval(uinput().title())

    @property
    def is_telemetry_enabled(self) -> bool:
        print(f"EP: isTelemetryEnabled", flush=True, end="")
        return eval(uinput().title())

    @property
    def language(self):
        print(f"EP: language", flush=True, end="")
        return uinput()

    @property
    def machine_id(self):
        print(f"EP: machineId", flush=True, end="")
        return uinput()

    @property
    def remote_name(self):
        print(f"EP: remoteName", flush=True, end="")
        return uinput()

    @property
    def session_id(self):
        print(f"EP: sessionId", flush=True, end="")
        return uinput()

    @property
    def shell(self):
        print(f"EP: shell", flush=True, end="")
        return uinput()

    @property
    def ui_kind(self):
        print(f"EP: uiKind", flush=True, end="")
        return uinput()

    @property
    def uri_scheme(self):
        print(f"EP: uriScheme", flush=True, end="")
        return uinput()


env = Env()



import os
import json
import time
import inspect
from .extension import Extension


def create_package(data, config):
    package_name = data["name"]
    package = {
        "name": package_name,
        "displayName": data.get("display_name", package_name),
        "version": data["version"],
        "engines": {"vscode": "^1.58.0"},
        "categories": ["Other"],
        "main": "./build/extension.js",
    }
    package.update(config)
    return package


extensions_json = {"recommendations": ["dbaeumer.vscode-eslint"]}

launch_json = {
    "version": "0.2.0",
    "configurations": [
        {
            "name": "Run Extension",
            "type": "extensionHost",
            "request": "launch",
            "args": ["--extensionDevelopmentPath=${workspaceFolder}"],
        },
        {
            "name": "Extension Tests",
            "type": "extensionHost",
            "request": "launch",
            "args": [
                "--extensionDevelopmentPath=${workspaceFolder}",
                "--extensionTestsPath=${workspaceFolder}/test/suite/index",
            ],
        },
    ],
}

pre_imports = """# Built using vscode-ext

import sys
"""


main_py = """
def ipc_main():
    globals()[sys.argv[1]]()

ipc_main()
"""


def build_py(functions):
    with open(inspect.getfile(functions[0]), "r") as f:
        imports = pre_imports + "".join([l for l in f.readlines() if not "build(" in l])
    imports += "\n"
    main = main_py
    code = imports + main
    return code


def build_js(name, events, commands, activity_bar_config=None):
    cwd = os.getcwd()
    python_path = os.path.join(cwd, "build", "extension.py").replace("\\", "\\\\")

    imports = ""
    directory, filename = os.path.split(inspect.getfile(build_py))
    with open(os.path.join(directory, "main.js"), "r") as f:
        imports += f.read()
    on_activate = events.get("activate")
    code_on_activate = "function activate(context) {\nlet globalStorage = {}\n"
    if on_activate:
        r = str(on_activate()).replace('"', '\\"')
        code_on_activate += f'console.log("{r}");\n'
    if activity_bar_config:
        html = activity_bar_config["html"].replace('"', '\\"')
        code_on_activate += (
            f'let html = "{html}"; let id = "{activity_bar_config["id"]}";\n'
        )
        code_on_activate += """
let thisProvider = {
  resolveWebviewView: function (thisWebview, thisWebviewContext, thisToken) {
    thisWebview.webview.options = { enableScript: true };
    thisWebview.webview.html = html;
  },
};
context.subscriptions.push(
  vscode.window.registerWebviewViewProvider(id, thisProvider)
);
            
            
"""
    for command in commands:
        code_on_activate += (
            f"let {command.name} = vscode.commands.registerCommand('{command.extension(name)}',"
            + "async function () {\n"
        )
        pyvar = "python" if os.name == "nt" else "python3"
        code_on_activate += (
            f'let funcName = "{command.func_name}"; let pyVar = "{pyvar}";'
        )
        code_on_activate += """
let py = spawn(pyVar, [pythonPath, funcName]);

py.stdout.on("data", (data) => {
    executeCommands(py, data, globalStorage);
});
py.stderr.on("data", (data) => {
    console.error(`An Error occurred in the python script: ${data}`);
});
"""
        code_on_activate += "});\n"
        code_on_activate += f"context.subscriptions.push({command.name});\n"

    code_on_activate += "}\n"

    on_deactivate = events.get("deactivate")
    code_on_deactivate = "function deactivate() {"
    if on_deactivate:
        r = str(on_deactivate()).replace('"', '\\"')
        code_on_activate += f'\nconsole.log("{r}");\n'
    code_on_deactivate += "}"
    main = code_on_activate + "\n" + code_on_deactivate
    exports = "module.exports = {activate,deactivate}"
    code = f"{imports}\n{main}\n\n{exports}"
    return code


def create_files(package, javascript, python, publish):
    cwd = os.getcwd()

    # ---- Static ----

    vscode_path = os.path.join(cwd, ".vscode")
    os.makedirs(vscode_path, exist_ok=True)
    os.chdir(vscode_path)

    with open("extensions.json", "w") as f:
        json.dump(extensions_json, f, indent=2)

    with open("launch.json", "w") as f:
        json.dump(launch_json, f, indent=2)

    # ---- Dynamic ----
    package_dir = os.path.join(cwd, "package.json")
    if os.path.isfile(package_dir):
        with open(package_dir, "r") as f:
            try:
                existing = json.load(f)
                existing.update(package)
            except json.decoder.JSONDecodeError:
                existing = package
    else:
        existing = package
    with open(package_dir, "w") as f:
        json.dump(existing, f, indent=2)

    build_path = os.path.join(cwd, "build")
    os.makedirs(build_path, exist_ok=True)
    os.chdir(build_path)
    with open("extension.js", "w") as f:
        f.write(javascript)

    with open("extension.py", "w") as f:
        f.write(python)
    os.chdir(cwd)

    if not os.path.isfile("requirements.txt"):
        with open("requirements.txt", "w") as f:
            f.write("vscode-ext")

    if publish:
        if not os.path.isfile("README.md"):
            with open("README.md", "w") as f:
                pass

        if not os.path.isfile("CHANGELOG.md"):
            with open("CHANGELOG.md", "w") as f:
                pass

        if not os.path.isfile(".vscodeignore"):
            with open(".vscodeignore", "w") as f:
                f.write(".vscode/**")


def build(extension: Extension, publish: bool = False, config: dict = None) -> None:
    """
    Builds the extension.
    """
    if config is None:
        config = {}
    if publish:
        if extension.publisher is None:
            config["publisher"] = input("Enter publisher name: ")
        else:
            config["publisher"] = extension.publisher

    print(f"\033[1;37;49mBuilding Extension {extension.name}...", "\033[0m")
    start = time.time()

    ext_data = extension.__dict__
    package_name = ext_data["name"]

    commands = []
    activation_events = []
    for command in ext_data.get("commands"):
        cmd = {"command": f"{package_name}.{command.name}", "title": command.title}
        if command.category is not None:
            cmd.update({"category": command.category})
        event = "onCommand:" + command.extension(package_name)
        commands.append(cmd)
        activation_events.append(event)

    package_config = config
    package_config.update(
        {
            "contributes": {
                "commands": commands,
            },
            "activationEvents": activation_events,
        }
    )

    if extension.keybindings:
        package_config["contributes"].update({"keybindings": extension.keybindings})

    if extension.activity_bar:
        package_config["contributes"]["viewsContainers"] = {
            "activitybar": [extension.activity_bar]
        }
        bar = extension.activity_bar
        webview = extension.activity_bar_webview
        view = {
            extension.activity_bar["id"]: [
                {
                    "id": f'{extension.name}.{bar["id"]}'
                    if not webview
                    else webview["id"],
                    "name": webview["title"]
                    if webview and webview["title"]
                    else bar["title"],
                }
            ]
        }
        if extension.activity_bar_webview:
            view[extension.activity_bar["id"]][0].update({"type": "webview"})
            package_config["activationEvents"].append(
                f"onView:{extension.activity_bar_webview['id']}"
            )
        if "views" in package_config["contributes"]:
            package_config["contributes"]["views"].update(view)
        else:
            package_config["contributes"]["views"] = view
    if extension.description:
        package_config["description"] = extension.description
    if extension.icon:
        package_config["icon"] = extension.icon
    if extension.repository:
        package_config["repository"] = extension.repository

    package = create_package(ext_data, package_config)
    javascript = build_js(
        package_name,
        ext_data["events"],
        ext_data["commands"],
        extension.activity_bar_webview,
    )
    python = build_py([c.func for c in ext_data["commands"]])
    create_files(package, javascript, python, publish)
    end = time.time()
    time_taken = round((end - start) * 1000, 2)
    print(f"\033[1;37;49mBuild completed successfully in {time_taken} ms!", "\033[0m")



import json

class ActivityBar:
    """
    Content settings for the activity bar.
    """

    def __init__(self, id: str, title: str, icon: str) -> None:
        self.id = id
        self.title = title
        self.icon = icon


class StaticWebview:
    """
    Content settings for a Static Webview.
    """

    def __init__(self, id: str, html: str, title: str = None) -> None:
        self.id = id
        self.html = html
        self.title = title


class InputBoxOptions:
    """
    Options to configure the behavior of the input box UI.
    """

    def __init__(
        self,
        title: str = None,
        password: bool = None,
        ignore_focus_out: bool = None,
        prompt: str = None,
        place_holder: str = None,
        value: str = None,
    ) -> None:
        self.title = title
        self.password = password
        self.ignoreFocusOut = ignore_focus_out
        self.prompt = prompt
        self.placeHolder = place_holder
        self.value = value


class QuickPickOptions:
    """
    Options to configure the behavior of the quick pick UI.
    """

    def __init__(
        self,
        title: str = None,
        can_pick_many: bool = None,
        ignore_focus_out: bool = None,
        match_on_description: bool = None,
        place_holder: str = None,
        match_on_detail: bool = None,
    ) -> None:
        self.title = title
        self.canPickMany = can_pick_many
        self.ignoreFocusOut = ignore_focus_out
        self.matchOnDescription = match_on_description
        self.placeHolder = place_holder
        self.matchOnDetail = match_on_detail


class QuickPickItem:
    """
    Content settings for a Quick Pick Item.
    """

    def __init__(
        self, label: str = None, detail: str = None, description: str = None, **options
    ) -> None:
        self.label = label
        self.detail = detail
        self.description = description
        self.__dict__.update(options)


class Undefined:
    """
    An instance of this class is returned everytime javascript returns undefined.
    """

    def __str__(self):
        return "undefined"

    def __bool__(self):
        return False

    def __eq__(self, other):
        return isinstance(other, self.__class__)


undefined = Undefined()


class Disposable:
    """
    Represents a type which can release resources, such as event listening or a timer.
    """

    def __init__(self, id):
        self.id = id

    def dispose(self):
        print(f"DI: {self.id}", flush=True, end="")


class Position:
    """
    Represents a line and character position, such as the position of the cursor.
    """

    def __init__(self, line: int, character: int):
        self.line = line
        self.character = character

    @staticmethod
    def from_dict(data: dict):
        pos = Position(0, 0)
        pos.__dict__.update(data)
        return pos

    def __eq__(self, other):
        return self.line == other.line and self.character == other.character

    def __sub__(self, other):
        return self.compare(other)

    def __lt__(self, other):
        return self.line < other.line or (
            self.line == other.line and self.character < other.character
        )

    def __le__(self, other):
        return self == other or self < other

    def __gt__(self, other):
        return not self <= other

    def __ge__(self, other):
        return not self < other

    def compare(self, other):
        return 1 if self > other else -1 if self < other else 0

    def __repr__(self):
        return f"{self.line}:{self.character}"


class Range:
    "A range represents an ordered pair of two positions. It is guaranteed that start.isBeforeOrEqual(end)"

    def __init__(self, start: Position, end: Position):
        self.start = start
        self.end = end

    @staticmethod
    def from_dict(data):
        return Range(Position.from_dict(data["start"]), Position.from_dict(data["end"]))

    @property
    def __dict__(self):
        return {"start": self.start.__dict__, "end": self.end.__dict__}

    @property
    def is_empty(self):
        return self.start == self.end

    @property
    def in_single_line(self):
        return self.start.line == self.end.line

    def __repr__(self):
        return f"[{self.start},{self.end}]"

    def __eq__(self, other):
        return self.start == other.start and self.end == other

    def __contains__(self, other):
        if isinstance(other, Position):
            return self.start <= other <= self.end
        else:
            return self.start <= other.start and self.end >= other.end

    def intersection(self, other):
        if self.end < other.start or other.end < self.start:
            return undefined

        start = max(self.start, other.start)
        end = min(self.end, other.end)

        return Range(start, end)

    def union(self, other):
        start = min(self.start, other.start)
        end = max(self.end, other.end)

        return Range(start, end)

class TextLine:
    """
    Represents a line of text, such as a line of source code.

    TextLine objects are immutable. When a document changes, previously retrieved lines will not represent the latest state.
    """
    def __init__(self, data):
        self.__dict__.update(apply_func_to_keys(data, camel_to_snake))

class TextDocument:
    """
    Represents a text document, such as a source file. Text documents have lines and knowledge about an underlying resource like a file.
    """

    def __init__(self, data):
        self.__dict__.update(data)

    def get_text(self, location: Range = None) -> str:
        """
        Get the text of this document. A substring can be retrieved by providing a range. The range will be adjusted.
        """
        if location is not None:
            if isinstance(location, Range):
                location = location.__dict__
            location = json.dumps(location)
            print(f"GT: {location}", flush=True, end="")
        else:
            print("GT", flush=True, end="")

        return json_input()

    def line_at(self, location) -> TextLine:
        """
        Returns a text line denoted by the line number or Position. Note that the returned object is not live and changes to the document are not reflected.
        
        Note: If location.line > TextDocument.line_count, the last line will be returned.
        """
        if isinstance(location, Position):
            location = location.line
        if location > self.line_count:
            location = self.line_count
 
        print(f'LA: {location}', flush=True, end="")
        return TextLine(json_input())

    # TODO: TextDocument methods

def uinput():
    res = input()
    if res.strip() == "undefined":
        return undefined
    else:
        return res


def json_input():
    res = uinput()
    if not res:
        return res
    try:
        return json.loads(res)
    except json.decoder.JSONDecodeError:
        return res

def camel_to_snake(text:str) -> str:
    return ''.join(['_'+i.lower() if i.isupper() else i for i in text]).lstrip('_')

def apply_func_to_keys(dictionary: dict, func) -> dict:
    new = {}
    for key in dictionary:
        if isinstance(dictionary[key], dict):
            new[func(key)] = apply_func_to_keys(dictionary[key], func)
        else:
            new[func(key)] = dictionary[key]
    return new


import json
from ._types import undefined


def uinput():
    res = input()
    if res.strip() == "undefined":
        return undefined
    else:
        return res


def json_input():
    res = uinput()
    if not res:
        return res
    try:
        return json.loads(res)
    except json.decoder.JSONDecodeError:
        return res

def camel_to_snake(text:str) -> str:
    return ''.join(['_'+i.lower() if i.isupper() else i for i in text]).lstrip('_')

def apply_func_to_keys(dictionary: dict, func) -> dict:
    new = {}
    for key in dictionary:
        if isinstance(dictionary[key], dict):
            new[func(key)] = apply_func_to_keys(dictionary[key], func)
        else:
            new[func(key)] = dictionary[key]
    return new